import pickle
import copy
import gc

import pytest

from registtro import RegistryProtocol, Registry, RegistryEvolver


@pytest.mark.parametrize("cls", (Registry, RegistryEvolver))
def test_protocol(cls):
    try:
        from typing import runtime_checkable
    except ImportError:
        pass
    else:
        assert issubclass(cls, RegistryProtocol)

    class Entry:
        pass

    entry_a = Entry()
    entry_b = Entry()
    self = cls()

    self = self.update({entry_a: 1})
    assert self.query(entry_a) == 1
    assert self.get(entry_b) is None

    self = self.update({entry_b: 2})
    assert self.query(entry_a) == 1
    assert self.query(entry_b) == 2

    assert len(self.to_dict()) == 2
    assert self.to_dict() == {entry_a: 1, entry_b: 2}


def test_registry_garbage_collection():
    class Entry:
        pass

    entry_a = Entry()
    entry_b = Entry()
    entry_c = Entry()
    registry_a = Registry({entry_a: 1})
    registry_b = registry_a.update({entry_b: 2})
    registry_c = registry_b.update({entry_c: 3})

    del entry_a
    gc.collect()

    assert registry_a.to_dict() == {}
    assert registry_b.to_dict() == {entry_b: 2}
    assert registry_c.to_dict() == {entry_b: 2, entry_c: 3}

    del entry_b
    gc.collect()

    assert registry_a.to_dict() == {}
    assert registry_b.to_dict() == {}
    assert registry_c.to_dict() == {entry_c: 3}

    del entry_c
    gc.collect()

    assert registry_a.to_dict() == {}
    assert registry_b.to_dict() == {}
    assert registry_c.to_dict() == {}


def test_evolver_garbage_collection():
    class Entry:
        pass

    entry_a = Entry()
    entry_b = Entry()
    entry_c = Entry()
    registry = Registry({entry_a: 1})

    evolver = registry.get_evolver()
    evolver.update({entry_b: 2, entry_c: 3})
    assert evolver.is_dirty()

    del entry_a, entry_b, entry_c
    gc.collect()

    assert len(evolver.to_dict()) == 2

    evolver.commit()
    gc.collect()

    assert not evolver.is_dirty()
    assert not evolver.to_dict()


def test_registry_evolver_roundtrip():
    class Entry:
        pass

    entry_a = Entry()
    entry_b = Entry()
    entry_c = Entry()
    registry = Registry({entry_a: 1, entry_b: 2})

    evolver = RegistryEvolver(registry)
    assert registry.to_dict() == evolver.to_dict()

    evolver.update({entry_c: 3})
    assert registry.to_dict() != evolver.to_dict()

    new_evolver = evolver.fork()
    evolver.reset()
    assert registry.to_dict() == evolver.to_dict()

    new_registry = new_evolver.get_registry()
    assert new_registry.to_dict() == new_evolver.to_dict()
    assert new_registry.to_dict() == {entry_a: 1, entry_b: 2, entry_c: 3}


@pytest.mark.parametrize("deep_copier", (copy.deepcopy, lambda s: pickle.loads(pickle.dumps(s))))
def test_deep_copy_and_pickle_registry(deep_copier):
    class _Entry:
        __name__ = __qualname__ = "_Entry"

        def __init__(self, name):
            self.name = name

        def __hash__(self):
            return hash(self.name)

        def __eq__(self, other):
            return other.name == self.name

    globals()[_Entry.__name__] = _Entry

    entry_a = _Entry("a")
    entry_b = _Entry("b")
    entry_c = _Entry("c")
    entry_d = _Entry("d")
    entries = entry_a, entry_b, entry_c
    registry = Registry({entry_a: 1, entry_b: 2, entry_c: 3, entry_d: 4})

    copied_entries, copied_registry = deep_copier((entries, registry))
    truth_dict = registry.to_dict()
    del truth_dict[entry_d]
    assert copied_registry.to_dict() == truth_dict


@pytest.mark.parametrize("deep_copier", (copy.deepcopy, lambda s: pickle.loads(pickle.dumps(s))))
def test_deep_copy_and_pickle_evolver(deep_copier):
    class _Entry:
        __name__ = __qualname__ = "_Entry"

        def __init__(self, name):
            self.name = name

        def __hash__(self):
            return hash(self.name)

        def __eq__(self, other):
            return other.name == self.name

    globals()[_Entry.__name__] = _Entry

    entry_a = _Entry("a")
    entry_b = _Entry("b")
    entry_c = _Entry("c")
    entry_d = _Entry("d")
    entries = (entry_a,)
    evolver = Registry({entry_a: 1, entry_d: 4}).get_evolver().update({entry_b: 2, entry_c: 3})

    copied_entries, copied_evolver = deep_copier((entries, evolver))
    assert len(copied_evolver.to_dict()) == 3
    truth_dict = evolver.to_dict()
    del truth_dict[entry_d]
    assert copied_evolver.to_dict() == truth_dict


def test_shallow_copy_registry():
    class Entry:
        pass

    entry_a = Entry()
    registry = Registry({entry_a: 1})

    assert copy.copy(registry) is registry


def test_shallow_copy_evolver():
    class Entry:
        pass

    entry_a = Entry()
    evolver = Registry({entry_a: 1}).get_evolver()
    evolver_copy = copy.copy(evolver)
    evolver_forked = evolver.fork()

    assert evolver_copy is not evolver is not evolver_forked
    assert evolver_copy.to_dict() == evolver.to_dict() == evolver_forked.to_dict()


if __name__ == "__main__":
    pytest.main()



from ._protocol import RegistryProtocol
from ._registry import Registry, RegistryEvolver
from ._exceptions import RegisttroException, EntryNotFoundError

__all__ = ["RegistryProtocol", "Registry", "RegistryEvolver", "RegisttroException", "EntryNotFoundError"]



from tippo import Dict, Union, Protocol, TypeVar, Mapping, runtime_checkable


_ET = TypeVar("_ET")
_VT = TypeVar("_VT")
_ST = TypeVar("_ST")


@runtime_checkable
class RegistryProtocol(Protocol[_ET, _VT]):
    """Protocol for registry-like interfaces."""

    def update(self, updates):
        # type: (_ST, Mapping[_ET, _VT]) -> _ST
        """Update entries."""

    def query(self, entry):
        # type: (_ET) -> _VT
        """
        Query value for entry.

        :raises EntryNotFoundError: Entry not in the registry.
        """

    def get(self, entry, fallback=None):
        # type: (_ET, Union[_VT, None]) -> Union[_VT, None]
        """Get value for entry, return fallback value if not in the registry."""

    def to_dict(self):
        # type: () -> Dict[_ET, _VT]
        """Convert to dictionary."""



import copy
import functools

import six
import pyrsistent
from tippo import AbstractSet, Dict, Union, Generic, TypeVar, Mapping, WeakSet, cast, ref
from basicco import runtime_final, generic_meta
from pyrsistent.typing import PMap, PMapEvolver

from ._exceptions import EntryNotFoundError


_ET = TypeVar("_ET")
_VT = TypeVar("_VT")
_ST = TypeVar("_ST")


class _RegistryMeta(runtime_final.FinalizedMeta, generic_meta.GenericMeta):
    pass


@runtime_final.final
class Registry(six.with_metaclass(_RegistryMeta, Generic[_ET, _VT])):
    """Immutable weak entry/strong value registry."""

    __slots__ = ("__weakref__", "__previous", "__registries", "__data")

    def __init__(self, initial=None):
        # type: (Union[Mapping[_ET, _VT], None]) -> None
        self.__previous = None  # type: Union[ref[Registry[_ET, _VT]], None]
        self.__registries = WeakSet({self})  # type: WeakSet[Registry[_ET, _VT]]
        self.__data = cast(PMapEvolver[ref[_ET], _VT], pyrsistent.pmap().evolver())
        if initial is not None:
            self.__initialize(initial)

    def __contains__(self, entry):
        # type: (object) -> bool
        return ref(entry) in self.__data.persistent()

    def __reduce__(self):
        return type(self), (self.to_dict(),)

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        try:
            deep_copy = memo[id(self)]
        except KeyError:
            deep_copy = memo[id(self)] = Registry(copy.deepcopy(self.to_dict(), memo))
        return deep_copy

    def __copy__(self):
        return self

    @staticmethod
    def __clean(registries, weak_key):
        # type: (AbstractSet[Registry[_ET, _VT]], ref[_ET]) -> None
        for registry in registries:
            del registry.__data[weak_key]

    def __initialize(self, initial):
        # type: (Mapping[_ET, _VT]) -> None
        temp_registry = self.update(initial)
        self.__registries = registries = temp_registry.__registries
        registries.clear()
        registries.add(self)
        self.__data = temp_registry.__data

    def update(self, updates):
        # type: (Mapping[_ET, _VT]) -> Registry[_ET, _VT]
        """Update entries."""
        if not updates:
            return self

        registry = Registry.__new__(Registry)
        registry.__previous = ref(self)
        registry.__registries = registries = WeakSet({registry})

        # Update weak references.
        weak_updates = {}
        for entry, value in updates.items():
            weak_key = ref(entry, functools.partial(Registry.__clean, registries))
            weak_updates[weak_key] = value
        if not weak_updates:
            return self

        # Update previous registries.
        previous = self  # type: Union[Registry[_ET, _VT], None]
        while previous is not None:
            previous.__registries.add(registry)
            if previous.__previous is None:
                break
            previous = previous.__previous()

        registry.__data = self.__data.persistent().update(weak_updates).evolver()

        return registry

    def query(self, entry):
        # type: (_ET) -> _VT
        """
        Query value for entry.

        :raises EntryNotFoundError: Entry not in the registry.
        """
        try:
            return self.__data[ref(entry)]
        except KeyError:
            exc = EntryNotFoundError(entry)
            six.raise_from(exc, None)
            raise exc

    def get(self, entry, fallback=None):
        # type: (_ET, Union[_VT, None]) -> Union[_VT, None]
        """Get value for entry, return fallback value if not in the registry."""
        try:
            return self.query(entry)
        except EntryNotFoundError:
            return fallback

    def to_dict(self):
        # type: () -> Dict[_ET, _VT]
        """Convert to dictionary."""
        to_dict = {}
        for weak_key, data in self.__data.persistent().items():
            entry = weak_key()
            if entry is not None:
                to_dict[entry] = data
        return to_dict

    def get_evolver(self):
        # type: () -> RegistryEvolver[_ET, _VT]
        """Get mutable evolver."""
        return RegistryEvolver(self)


@runtime_final.final
class RegistryEvolver(six.with_metaclass(_RegistryMeta, Generic[_ET, _VT])):
    """Mutable registry evolver."""

    __slots__ = ("__weakref__", "__registry", "__updates")

    def __init__(self, registry=None):
        # type: (Union[Registry, None]) -> None
        if registry is None:
            registry = Registry()
        self.__registry = registry  # type: Registry[_ET, _VT]
        self.__updates = pyrsistent.pmap()  # type: PMap[_ET, _VT]

    def __contains__(self, entry):
        # type: (object) -> bool
        return entry in self.__updates or entry in self.__registry

    def __reduce__(self):
        return _evolver_reducer, (self.__registry, self.__updates)

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        try:
            deep_copy = memo[id(self)]
        except KeyError:
            deep_copy = memo[id(self)] = RegistryEvolver.__new__(RegistryEvolver)
            deep_copy_args_a = self.__registry, memo
            deep_copy.__registry = copy.deepcopy(*deep_copy_args_a)
            deep_copy_args_b = self.__updates, memo
            deep_copy.__updates = copy.deepcopy(*deep_copy_args_b)
        return deep_copy

    def __copy__(self):
        return self.fork()

    def update(self, updates):
        # type: (Mapping[_ET, _VT]) -> RegistryEvolver[_ET, _VT]
        """Update entries."""
        self.__updates = self.__updates.update(updates)
        return self

    def query(self, entry):
        # type: (_ET) -> _VT
        """
        Query value for entry.

        :raises EntryNotFoundError: Entry not in the registry.
        """
        try:
            return self.__updates[entry]
        except KeyError:
            try:
                return self.__registry.query(entry)
            except EntryNotFoundError:
                exc = EntryNotFoundError(entry)
                six.raise_from(exc, None)
                raise exc

    def get(self, entry, fallback=None):
        # type: (_ET, Union[_VT, None]) -> Union[_VT, None]
        """Get value for entry, return fallback value if not in the registry."""
        try:
            return self.query(entry)
        except EntryNotFoundError:
            return fallback

    def to_dict(self):
        # type: () -> Dict[_ET, _VT]
        """Convert to dictionary."""
        return self.get_registry().to_dict()

    def get_registry(self):
        # type: () -> Registry[_ET, _VT]
        """Get immutable registry."""
        return self.__registry.update(self.__updates)

    def fork(self):
        # type: () -> RegistryEvolver[_ET, _VT]
        """Fork into a new mutable evolver."""
        evolver = RegistryEvolver.__new__(RegistryEvolver)
        evolver.__registry = self.__registry
        evolver.__updates = self.__updates
        return evolver

    def is_dirty(self):
        # type: () -> bool
        """Whether has updates that were not committed."""
        return bool(self.__updates)

    def reset(self):
        """Reset updates to last commit."""
        self.__updates = pyrsistent.pmap()

    def commit(self):
        """Commit updates."""
        self.__registry = self.__registry.update(self.__updates)
        self.__updates = pyrsistent.pmap()

    @property
    def updates(self):
        # type: () -> PMap[_ET, _VT]
        """Updates."""
        return self.__updates


def _evolver_reducer(registry, updates):
    # type: (Registry[_ET, _VT], Mapping[_ET, _VT]) -> RegistryEvolver[_ET, _VT]
    evolver = RegistryEvolver(registry)  # type: RegistryEvolver[_ET, _VT]
    evolver.update(updates)
    return evolver
